module Klastera::Concerns::User
  extend ActiveSupport::Concern

  included do

    CLUSTER_ROOT = 'root'.freeze
    CLUSTER_ADMIN = 'admin'.freeze
    CLUSTER_USER = 'user'.freeze

    CLUSTER_ROLES = [ CLUSTER_ROOT, CLUSTER_ADMIN, CLUSTER_USER ].freeze

    has_many :cluster_users, inverse_of: :user, class_name: ::ClusterUser.to_s
    accepts_nested_attributes_for :cluster_users, reject_if: :all_blank, allow_destroy: true
    validates :cluster_role, presence: true, if: -> { organization.present? && organization.is_in_cluster_mode? }
    validates :cluster_role, inclusion: { in: CLUSTER_ROLES , message: I18n.t('klastera.clusters.wrong_option') }, if: -> { cluster_role.present? }
    validate :at_least_one_role, if: -> { use_cluster? && try(:need_cluster_assignation?) }

    def use_cluster?; cluster_role.present?; end

    def is_a_cluster_root?; cluster_role == CLUSTER_ROOT; end
    def is_a_cluster_admin?; cluster_role == CLUSTER_ADMIN; end
    def is_a_cluster_user?; cluster_role == CLUSTER_USER; end

    def need_cluster_assignation?
      organization.present? && organization.required_suborganization_mode? && is_a_cluster_user?
    end

    def can_admin_clusters?
      organization.present? && organization.is_in_cluster_mode? && ( is_a_cluster_admin? || is_a_cluster_root? )
    end

    #
    # It tells if this user should see what they cluster role or organization dictates
    # Adminers and users from show cluster modes skip cluster clause
    def cannot_skip_cluster_clause?
      ! ( is_a_cluster_admin? || is_a_cluster_root? || ( organization.present? && organization.optional_suborganization_mode? && cluster_users.blank? ) )
    end

    #
    # We will try to create a cluster_user record whether
    # the organization is using force mode (if cluster_id isn't present, it will raise a validation error) 
    # or
    # cluster_id is present (on show/use mode organizations).
    #
    def try_to_clusterify!(role,cluster_ids)
      if organization.present? && organization.is_in_cluster_mode?
        self.cluster_role = role
        if ( organization.present? && organization.required_suborganization_mode? ) || cluster_ids.present?
          timestamp = DateTime.now.to_i
          nested_attributes = {}
          cluster_ids.each do |cluster_id|
            nested_attributes["#{cluster_id}_#{timestamp}"] = {
              cluster_id: cluster_id,
              _destroy: false
            }
          end
          self.cluster_users_attributes = nested_attributes
        end
      end
    end

    private

    def at_least_one_role
      if cluster_users.none?
        errors.add(:base, I18n.t('klastera.user.you_need_add_at_least_one_cluster'))
      end
    end

  end
  
  module ClassMethods
    def cluster_params
      [ :cluster_role, { cluster_users_attributes: [:id, :cluster_id, :_destroy] } ]
    end
  end
end